package liber

import (
	"bytes"
	"encoding/binary"

	"github.com/syndtr/goleveldb/leveldb"
	ldbfilter "github.com/syndtr/goleveldb/leveldb/filter"
	ldbiter "github.com/syndtr/goleveldb/leveldb/iterator"
	ldbopt "github.com/syndtr/goleveldb/leveldb/opt"
	ldbutil "github.com/syndtr/goleveldb/leveldb/util"
)

var (
	ldbBookBucket     = []byte("ebook")
	ldbFileBucket     = []byte("file")
	ldbBookFileBucket = []byte("ebook_file")
)

type leveldbDB struct {
	ldb *leveldb.DB
}

func newLevelDBBackend(path string) (*leveldbDB, error) {
	// Use 128MB of cache and a small Bloom filter.
	opts := &ldbopt.Options{
		Filter:             ldbfilter.NewBloomFilter(10),
		BlockCacheCapacity: 1 << 28, // 256MB
	}

	ldb, err := leveldb.OpenFile(path, opts)
	if err != nil {
		return nil, err
	}

	l := &leveldbDB{ldb}
	if err := l.runMigrations(leveldbMigrations); err != nil {
		ldb.Close()
		return nil, err
	}

	return l, nil
}

func (db *leveldbDB) Close() {
	db.ldb.Close()
}

var (
	schemaVersionKey  = []byte("_liber_schema_version")
	leveldbMigrations []databaseMigration
)

func (db *leveldbDB) getSchemaVersion() uint64 {
	data, err := db.ldb.Get(schemaVersionKey, nil)
	if err != nil {
		return 0
	}
	return binary.LittleEndian.Uint64(data)
}

func (db *leveldbDB) setSchemaVersion(v uint64) {
	var b [8]byte
	binary.LittleEndian.PutUint64(b[:], v)
	db.ldb.Put(schemaVersionKey, b[:], nil)
}

type databaseMigration struct {
	version uint64
	run     func(db *leveldb.DB) error
}

func (db *leveldbDB) runMigrations(migrations []databaseMigration) error {
	version := db.getSchemaVersion()
	for _, m := range migrations {
		if m.version < version {
			continue
		}
		if err := m.run(db.ldb); err != nil {
			return err
		}
		version = m.version
		db.setSchemaVersion(version)
	}
	return nil
}

func (db *leveldbDB) get(bucket, key []byte, obj interface{}) error {
	data, err := db.ldb.Get(levelDBBucketToKey(bucket, key), nil)
	if err != nil {
		return err
	}
	return gobDec(data, obj)
}

func (db *leveldbDB) GetBook(bookid BookId) (*Book, error) {
	var b Book
	if err := db.get(ldbBookBucket, bookid.Key(), &b); err != nil {
		return nil, err
	}
	return &b, nil
}

func (db *leveldbDB) GetFile(path string) (*File, error) {
	var f File
	if err := db.get(ldbFileBucket, []byte(path), &f); err != nil {
		return nil, err
	}
	return &f, nil
}

func (db *leveldbDB) GetBookFiles(bookid BookId) ([]*File, error) {
	start, end := keyRange(levelDBBucketToKey(ldbBookFileBucket, bookid.Key()))
	it := db.ldb.NewIterator(
		&ldbutil.Range{
			Start: start,
			Limit: end,
		},
		nil,
	)
	defer it.Release()

	var out []*File
	for it.Next() {
		var filepath string
		if gobDec(it.Value(), &filepath) == nil {
			if file, err := db.GetFile(filepath); err == nil {
				out = append(out, file)
			}
		}
	}
	if err := it.Error(); err != nil {
		return nil, err
	}
	return out, nil
}

func (db *leveldbDB) put(bucket, key []byte, obj interface{}) error {
	data, err := gobEnc(obj)
	if err != nil {
		return err
	}
	if bucket != nil {
		key = levelDBBucketToKey(bucket, key)
	}
	return db.ldb.Put(key, data, nil)
}

func (db *leveldbDB) PutBook(b *Book) error {
	return db.put(ldbBookBucket, b.Id.Key(), b)
}

func (db *leveldbDB) PutFile(f *File) error {
	if err := db.put(ldbFileBucket, []byte(f.Path), f); err != nil {
		return err
	}
	if !f.Error {
		return db.put(ldbBookFileBucket, levelDBFileBookKey(f.Path, f.Id), f.Path)
	}
	return nil
}

func (db *leveldbDB) delete(bucket, key []byte) error {
	return db.ldb.Delete(levelDBBucketToKey(bucket, key), nil)
}

func (db *leveldbDB) DeleteBook(bookid BookId) error {
	return db.delete(ldbBookBucket, bookid.Key())
}

func (db *leveldbDB) DeleteFile(path string) error {
	f, err := db.GetFile(path)
	if err != nil {
		return nil
	}

	db.delete(ldbFileBucket, []byte(path))
	db.delete(ldbBookFileBucket, levelDBFileBookKey(path, f.Id))

	// Delete the book if there are no files left.
	if files, err := db.GetBookFiles(f.Id); err == nil && len(files) == 0 {
		db.DeleteBook(f.Id)
	}

	return nil
}

func (db *leveldbDB) scanBucket(bucket []byte) ldbiter.Iterator {
	start, end := keyRange(bucket)
	return db.ldb.NewIterator(&ldbutil.Range{
		Start: start,
		Limit: end,
	}, &ldbopt.ReadOptions{
		DontFillCache: true,
	})
}

func (db *leveldbDB) ScanBooks() <-chan *Book {
	ch := make(chan *Book, 100)
	go func() {
		defer close(ch)
		iter := db.scanBucket(ldbBookBucket)
		defer iter.Release()
		for iter.Next() {
			var b Book
			if err := gobDec(iter.Value(), &b); err != nil {
				continue
			}
			ch <- &b
		}
	}()
	return ch
}

func (db *leveldbDB) ScanFiles() <-chan *File {
	ch := make(chan *File, 100)
	go func() {
		defer close(ch)
		iter := db.scanBucket(ldbFileBucket)
		defer iter.Release()
		for iter.Next() {
			var f File
			if err := gobDec(iter.Value(), &f); err != nil {
				continue
			}
			ch <- &f
		}
	}()
	return ch
}

var ldbKeySeparator = []byte{'/'}

func levelDBBucketToKey(bucket, key []byte) []byte {
	return bytes.Join([][]byte{bucket, key}, ldbKeySeparator)
}

func levelDBFileBookKey(path string, bookid BookId) []byte {
	return bytes.Join([][]byte{bookid.Key(), []byte(path)}, ldbKeySeparator)
}

func keyRange(prefix []byte) ([]byte, []byte) {
	start := make([]byte, len(prefix)+1)
	end := make([]byte, len(prefix)+1)
	copy(start, prefix)
	copy(end, prefix)
	start[len(prefix)] = ldbKeySeparator[0]
	end[len(prefix)] = ldbKeySeparator[0] + 1
	return start, end
}
