package liber

import (
	"bytes"
	cryptorand "crypto/rand"
	"encoding/binary"
	"encoding/gob"
	"errors"
	"fmt"
	"io"
	"log"
	"math/rand"
	"os"
	"path/filepath"
	"strconv"
	"time"

	"github.com/blevesearch/bleve"
	"github.com/blevesearch/bleve/analysis/analyzer/custom"
	"github.com/blevesearch/bleve/analysis/analyzer/keyword"
	"github.com/blevesearch/bleve/analysis/analyzer/simple"
	"github.com/blevesearch/bleve/analysis/analyzer/standard"
	"github.com/blevesearch/bleve/analysis/lang/en"
	"github.com/blevesearch/bleve/analysis/token/edgengram"
	"github.com/blevesearch/bleve/analysis/token/lowercase"
	"github.com/blevesearch/bleve/analysis/tokenizer/unicode"
	blevegoleveldb "github.com/blevesearch/bleve/index/store/goleveldb"
	blevemapping "github.com/blevesearch/bleve/mapping"
	blevequery "github.com/blevesearch/bleve/search/query"
)

// BookId is a unique identifier assigned to each book.
type BookId uint64

func (id BookId) String() string {
	return strconv.FormatUint(uint64(id), 10)
}

// Key returns a binary-encoded version of the identifier.
func (id BookId) Key() []byte {
	var b [8]byte
	binary.LittleEndian.PutUint64(b[:], uint64(id))
	return b[:]
}

// NewID creates a new random unique book identifier.
func NewID() BookId {
	return BookId(rand.Int63())
}

// ParseID parses a string with a hex-encoded representation of a
// book identifier.
func ParseID(s string) BookId {
	id, _ := strconv.ParseUint(s, 10, 64)
	return BookId(id)
}

// ParseBinaryID parses a binary-encoded book identifier.
func ParseBinaryID(b []byte) BookId {
	return BookId(binary.LittleEndian.Uint64(b))
}

// Book holds information about a book.
type Book struct {
	Id        BookId
	CoverPath string
	Metadata  *Metadata
}

func (b *Book) String() string {
	return fmt.Sprintf("%s (%s)", b.Metadata.String(), b.Id.String())
}

// File holds information about a file that is archived in our storage.
type File struct {
	Path     string
	FileType string
	Mtime    time.Time
	Size     int64
	Error    bool
	Id       BookId
}

// HasChanged returns true if the file metadata has changed compared
// to what is passed as argument.
func (f *File) HasChanged(info os.FileInfo) bool {
	return !info.ModTime().Equal(f.Mtime) || info.Size() != f.Size
}

func init() {
	// Seed the RNG to a random value.
	var seed int64
	binary.Read(cryptorand.Reader, binary.LittleEndian, &seed)
	rand.Seed(seed)
}

// The structure that gets actually indexed.
type flatBook struct {
	Title       string   `json:"title"`
	Author      []string `json:"author"`
	Description string   `json:"description"`
	ISBN        []string `json:"isbn"`
	Unique      []string `json:"_unique"`
	Suggest     []string `json:"_suggest"`
}

func (f *flatBook) Type() string {
	return "ebook"
}

func flatten(book *Book) *flatBook {
	suggest := []string{book.Metadata.Title}
	if len(book.Metadata.Creator) > 0 {
		suggest = append(suggest, book.Metadata.Creator...)
	}
	return &flatBook{
		Title:       book.Metadata.Title,
		Author:      book.Metadata.Creator,
		Description: book.Metadata.Description,
		ISBN:        book.Metadata.ISBN,
		Unique:      book.Metadata.Uniques(),
		Suggest:     suggest,
	}
}

var defaultTextAnalyzer = standard.Name

func metadataDocumentMapping() *blevemapping.DocumentMapping {
	md := bleve.NewDocumentStaticMapping()

	textFieldMapping := bleve.NewTextFieldMapping()
	textFieldMapping.Store = false
	textFieldMapping.Analyzer = defaultTextAnalyzer

	authorFieldMapping := bleve.NewTextFieldMapping()
	authorFieldMapping.Store = false
	authorFieldMapping.Analyzer = simple.Name

	keywordFieldMapping := bleve.NewTextFieldMapping()
	keywordFieldMapping.Store = false
	keywordFieldMapping.Analyzer = keyword.Name
	keywordFieldMapping.IncludeInAll = false

	suggestFieldMapping := bleve.NewTextFieldMapping()
	suggestFieldMapping.Store = false
	suggestFieldMapping.Analyzer = "edgeNgram"
	suggestFieldMapping.IncludeTermVectors = false
	suggestFieldMapping.IncludeInAll = false

	md.AddFieldMappingsAt("title", textFieldMapping)
	md.AddFieldMappingsAt("author", authorFieldMapping)
	md.AddFieldMappingsAt("description", textFieldMapping)
	md.AddFieldMappingsAt("isbn", keywordFieldMapping)
	md.AddFieldMappingsAt("_unique", keywordFieldMapping)
	md.AddFieldMappingsAt("_suggest", suggestFieldMapping)

	return md
}

func defaultIndexMapping() blevemapping.IndexMapping {
	i := bleve.NewIndexMapping()

	err := i.AddCustomTokenFilter("edgeNgram325",
		map[string]interface{}{
			"type": edgengram.Name,
			"min":  3.0,
			"max":  25.0,
		})
	if err != nil {
		log.Fatal("creating token filter edgeNgram325:", err)
	}

	err = i.AddCustomAnalyzer("edgeNgram",
		map[string]interface{}{
			"type":      custom.Name,
			"tokenizer": unicode.Name,
			"token_filters": []interface{}{
				en.PossessiveName,
				lowercase.Name,
				en.StopName,
				"edgeNgram325",
			},
		})
	if err != nil {
		log.Fatal("creating custom analyzer edgeNgram:", err)
	}

	i.AddDocumentMapping("ebook", metadataDocumentMapping())

	i.DefaultAnalyzer = defaultTextAnalyzer
	i.DefaultType = "ebook"
	return i
}

type backend interface {
	Close()

	GetBook(BookId) (*Book, error)
	GetFile(string) (*File, error)
	GetBookFiles(BookId) ([]*File, error)
	PutBook(*Book) error
	PutFile(*File) error
	DeleteBook(BookId) error
	DeleteFile(string) error

	ScanBooks() <-chan *Book
	ScanFiles() <-chan *File
}

// Database binds together a database of books/files with a searchable
// full-text index, making sure that updates are synchronized.
type Database struct {
	backend

	path  string
	index bleve.Index
}

// NewDb creates a new Database at the specified location. The only
// supported backendType at the moment is 'leveldb'.
func NewDb(backendType, path string) (*Database, error) {
	// Make sure that path exists.
	if _, err := os.Stat(path); err != nil {
		if err := os.Mkdir(path, 0700); err != nil {
			return nil, err
		}
	}

	var b backend
	switch backendType {
	case "leveldb":
		ldb, err := newLevelDBBackend(path)
		if err != nil {
			return nil, err
		}
		b = ldb
	default:
		return nil, errors.New("unknown backend")
	}

	d := &Database{backend: b, path: path}
	// Initialize the index.
	if err := d.setupIndex(filepath.Join(path, "index")); err != nil {
		b.Close()
		return nil, err
	}
	return d, nil
}

func (db *Database) setupIndex(path string) error {
	var err error
	if _, serr := os.Stat(path); serr == nil {
		db.index, err = bleve.Open(path)
	} else {
		// Create a new Bleve index, backed by goleveldb.
		db.index, err = bleve.NewUsing(path, defaultIndexMapping(), bleve.Config.DefaultIndexType, blevegoleveldb.Name, map[string]interface{}{
			"create_if_missing":         true,
			"write_buffer_size":         2 << 25,
			"lru_cache_capacity":        2 << 27,
			"bloom_filter_bits_per_key": 10,
		})
	}
	if err != nil {
		return err
	}
	return nil
}

// Close the database, the index, and all associated resources.
func (db *Database) Close() {
	db.index.Close()
	db.backend.Close()
}

// PutBook stores a new book in the database (or updates an existing
// one).
func (db *Database) PutBook(b *Book) error {
	if err := db.backend.PutBook(b); err != nil {
		return err
	}
	return db.index.Index(b.Id.String(), flatten(b))
}

// DeleteBook removes a book from the database.
func (db *Database) DeleteBook(bookid BookId) error {
	db.backend.DeleteBook(bookid)
	return db.index.Delete(bookid.String())
}

func writeBytes(w io.Writer, b []byte) error {
	binary.Write(w, binary.LittleEndian, uint32(len(b)))
	_, err := w.Write(b)
	return err
}

func readBytes(r io.Reader) ([]byte, error) {
	var szb [4]byte
	n, err := io.ReadFull(r, szb[:])
	if err != nil {
		return nil, err
	}
	if n != 4 {
		return nil, errors.New("short header read")
	}

	sz := binary.LittleEndian.Uint32(szb[:])
	b := make([]byte, sz)
	n, err = io.ReadFull(r, b)
	if n != int(sz) {
		return nil, errors.New("short data read")
	}
	return b, err
}

var (
	dumpTypeBook = []byte("book")
	dumpTypeFile = []byte("file")
)

func gobEnc(obj interface{}) ([]byte, error) {
	var buf bytes.Buffer
	if err := gob.NewEncoder(&buf).Encode(obj); err != nil {
		return nil, err
	}
	return buf.Bytes(), nil
}

func gobDec(data []byte, obj interface{}) error {
	return gob.NewDecoder(bytes.NewReader(data)).Decode(obj)
}

func dumpBook(w io.Writer, b *Book) {
	writeBytes(w, dumpTypeBook)
	data, _ := gobEnc(b)
	writeBytes(w, data)
}

func dumpFile(w io.Writer, f *File) {
	writeBytes(w, dumpTypeFile)
	data, _ := gobEnc(f)
	writeBytes(w, data)
}

// Dump the contents of the database to a Writer.
func (db *Database) Dump(w io.Writer) error {
	var count int
	for b := range db.backend.ScanBooks() {
		dumpBook(w, b)
		if files, err := db.backend.GetBookFiles(b.Id); err == nil {
			for _, f := range files {
				dumpFile(w, f)
			}
		}
		count++
	}
	log.Printf("dumped %d entries from the database", count)
	return nil
}

// Restore a backup to the current database (assuming it is empty).
func (db *Database) Restore(r io.Reader) error {
	var count int
	for {
		objType, err := readBytes(r)
		if err == io.EOF {
			break
		} else if err != nil {
			return err
		}

		objData, err := readBytes(r)
		if err == io.EOF {
			return errors.New("unexpected eof")
		} else if err != nil {
			return err
		}

		switch {
		case bytes.Equal(objType, dumpTypeBook):
			var b Book
			if err := gobDec(objData, &b); err != nil {
				return err
			}
			if err := db.PutBook(&b); err != nil {
				return err
			}
		case bytes.Equal(objType, dumpTypeFile):
			var f File
			if err := gobDec(objData, &f); err != nil {
				return err
			}
			if err := db.PutFile(&f); err != nil {
				return err
			}
		default:
			continue
		}
		count++
		if count%1000 == 0 {
			log.Printf("restored %d entries", count)
		}
	}
	log.Printf("restore complete (%d entries)", count)
	return db.Reindex()
}

// Reindex the entire database. This is an administrative operation,
// to be performed after an incompatible index schema change. It will
// delete the existing index and re-create it from scratch.
func (db *Database) Reindex() error {
	// Close the index, delete it, and re-open it.
	db.index.Close()

	indexPath := filepath.Join(db.path, "index")
	if err := os.RemoveAll(indexPath); err != nil {
		return err
	}
	if err := db.setupIndex(indexPath); err != nil {
		return err
	}

	// Scan the database and re-index everything. Use batch
	// indexing to achieve decent performance.
	i := 0
	start := time.Now()
	batch := db.index.NewBatch()
	if err := db.onAllBooks(func(book *Book) error {
		batch.Index(book.Id.String(), flatten(book))
		i++
		if i%100 == 0 {
			if err := db.index.Batch(batch); err != nil {
				return err
			}
			batch = db.index.NewBatch()
		}
		return nil
	}); err != nil {
		return err
	}

	// Flush the batch indexer if not empty.
	if err := db.index.Batch(batch); err != nil {
		return err
	}

	log.Printf("re-indexing complete, %d documents in %g seconds", i, time.Since(start).Seconds())
	return nil
}

// Call a function for all books in the database.
func (db *Database) onAllBooks(f func(*Book) error) error {
	for book := range db.ScanBooks() {
		if err := f(book); err != nil {
			return err
		}
	}
	return nil
}

// SearchResult holds the result of a search query. NumResults refers
// to the total number of results, regardless of pagination.
type SearchResult struct {
	Results    []*Book
	NumResults int
}

func (db *Database) doSearch(query blevequery.Query, offset, limit int) (*SearchResult, error) {
	req := bleve.NewSearchRequestOptions(query, limit, offset, false)
	result, err := db.index.Search(req)
	if err != nil {
		return nil, err
	}

	sr := SearchResult{NumResults: int(result.Total)}
	for _, r := range result.Hits {
		if book, err := db.GetBook(ParseID(r.ID)); err == nil {
			sr.Results = append(sr.Results, book)
		}
	}
	return &sr, nil
}

// Search the database with a query string.
func (db *Database) Search(queryStr string, offset, limit int) (*SearchResult, error) {
	return db.doSearch(bleve.NewQueryStringQuery(queryStr), offset, limit)
}

// Suggest runs a fuzzy search for a term, for autocompletion.
func (db *Database) Suggest(term string) (*SearchResult, error) {
	query := bleve.NewTermQuery(term)
	query.SetField("_suggest")
	return db.doSearch(query, 0, 20)
}

// Find a book matching the given metadata, if possible.
func (db *Database) Find(uniqueIds []string) (*Book, error) {
	var queries []blevequery.Query
	var query blevequery.Query

	for _, u := range uniqueIds {
		q := bleve.NewTermQuery(u)
		q.SetField("_unique")
		queries = append(queries, q)
	}
	if len(queries) > 0 {
		query = bleve.NewDisjunctionQuery(queries...)
	} else {
		query = queries[0]
	}

	search := bleve.NewSearchRequest(query)
	result, err := db.index.Search(search)
	if err != nil {
		return nil, err
	}
	if len(result.Hits) == 0 {
		return nil, errors.New("no matches found")
	}

	return db.GetBook(ParseID(result.Hits[0].ID))
}
